unit mmtimer;

interface
uses classes,windows,mmsystem;

const
  TIME_KILL_SYNCHRONOUS  = $0100;

type
  TMultiMediaTimer = class(TObject)
  private
    hmmTImer: THandle;
    fEnabled: boolean;
    fInterval: double;
    fOnTimer: TNotifyEvent;
    procedure SetEnabled(const Value: boolean);
    procedure SetInterval(const Value: double);
    procedure SetOnTimer(const Value: TNotifyEvent);
    procedure InitTimer;
  public
    timeCaps: TTimeCaps;
    function GetTicks: double;
    constructor Create;
    destructor Destroy;override;
  published
    property OnTimer: TNotifyEvent read fOnTimer write SetOnTimer;
    property Enabled: boolean read fEnabled write SetEnabled;
    property Interval: double read fInterval write SetInterval; // seconds
  end;


  TRealTimer = class(TThread) // the TRealTimer is an attempt to get realtime capabilities into windows
   private
    IsWaitableTimerEnabled,
    IsActive,
    fEnabled: boolean;
    fInterval: double;
    fOnTimer: TNotifyEvent;
    hTimer: THandle;
    procedure SetEnabled(const Value: boolean);
    procedure SetInterval(const Value: double);
    function GetTicks: double;
   protected
    procedure Execute;override;
   public
    CpuFreq: double;
    property Terminated;
    property Enabled: boolean read fEnabled write SetEnabled;
    property Interval: double read fInterval write SetInterval; // seconds
    property OnTimer: TNotifyEvent read fOnTimer write fOnTimer;
    constructor Create;
    destructor Destroy;override;
  end;

  procedure ReaDTimeStampCounter(var n: int64);
  
implementation

uses forms;

procedure ReaDTimeStampCounter(var n: int64);
asm
  mov ECX,eax
  db $0f,$31 //RDTSC
  mov [ecx],eax
  mov [ecx+4],edx
end;

procedure PrintLastError;
var lpMsgBuf: PChar;
begin
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER or FORMAT_MESSAGE_FROM_SYSTEM,nil,GetLastError,0,@lpMsgBuf,0,nil);
  Application.MessageBox(lpMsgBuf,'error',0);
  LocalFree(cardinal(lpMsgBuf));
end;

{ TMultiMediaTimer }

constructor TMultiMediaTimer.Create;

begin
  OnTimer := nil;
  Interval := 1;
  Enabled := false;
  hmmTImer:=0;

  timeCaps.wPeriodMin:=1000000;
  if TIMERR_NOERROR=timeGetDevCaps(@timeCaps,sizeof(timeCaps)) then begin
    timeBeginPeriod(timeCaps.wPeriodMin);
  end else
    timeCaps.wPeriodMin:=1000000;
end;

destructor TMultiMediaTimer.Destroy;
begin
  if hmmTImer<> 0 then begin
    timeKillEvent(hmmTImer);
    hmmTImer:=0;
  end;
  if timeCaps.wPeriodMin<1000000 then
    timeEndPeriod(timeCaps.wPeriodMin);
  inherited;
end;

//void (CALLBACK)(UINT uTimerID, UINT uMsg, DWORD_PTR dwUser, DWORD_PTR dw1, DWORD_PTR dw2);
procedure mmTImerCallback(uTimerID: cardinal;uMsg: cardinal; mmT: TMultiMediaTimer; dw1,dw2: Pointer); stdcall;
begin
  if Assigned(mmT.fOnTimer) then
    mmT.fOnTimer(mmT)
end;

function TMultiMediaTimer.GetTicks: double;
begin
  result:=GetTickCount*1000;
end;

procedure TMultiMediaTimer.InitTimer;
var dt: cardinal;
begin
  if hmmTImer<> 0 then begin
    timeKillEvent(hmmTImer);
    hmmTImer:=0;
  end;

  if not fEnabled then exit;
  if fInterval*1000 < timeCaps.wPeriodMin then exit;
  if fInterval*1000 > timeCaps.wPeriodMax then exit;
  if not Assigned(fOnTimer) then exit;

  dt := Round(Interval*1000);
  if dt<timeCaps.wPeriodMin then dt:=timeCaps.wPeriodMin;
  hmmTImer := timeSetEvent(dt,timeCaps.wPeriodMin,@mmTImerCallback,cardinal(self),TIME_ONESHOT or TIME_KILL_SYNCHRONOUS);
end;

procedure TMultiMediaTimer.SetEnabled(const Value: boolean);
begin
  fEnabled := Value;
  InitTimer;
end;

procedure TMultiMediaTimer.SetInterval(const Value: double);
begin
  fInterval := Value;
  InitTimer;
end;

procedure TMultiMediaTimer.SetOnTimer(const Value: TNotifyEvent);
begin
  fOnTimer := Value;
  InitTimer;
end;


//______________________________________________________________________________

{ TRealTimer }
constructor TRealTimer.Create;
var n1,n2: int64; t1: cardinal;
begin
  inherited Create(true);
  hTimer:= CreateWaitableTimer(nil,true,nil); // timer of 100ns accuracy (theory), not available under win95
  fEnabled := false;
  fInterval := 0;
  fOnTimer := nil;
  Priority := tpTimeCritical;

  try
    t1:=GetTickCount;
    ReaDTimeStampCounter(n1);
    Sleep(200);
    t1:=GetTickCount-t1;
    ReaDTimeStampCounter(n2);
    CpuFreq := 1000*(n2-n1)/t1;
  except
    CpuFreq := 0;
  end;
end;

destructor TRealTimer.Destroy;
begin
  Enabled := false; // stop the timer
  Terminate;  // set terminated true
  Resume;  // exit the thread loop
  if hTimer<>0 then CloseHandle(hTimer);
  inherited;
end;




procedure TRealTimer.Execute;
var t: int64; dt,d: double;
begin
//  Priority := tpHighest;
  Priority := tpTimeCritical;

  while not Terminated do begin
    IsActive:= true;
    dt:=$7fffffff;
    while fEnabled and (fInterval>0) do begin
      d:= GetTicks - dt;
      if (d>0) then begin
        if (d<fInterval) then d:= fInterval-d
        else d:=0.001;
      end else //}
      d:=fInterval;
      t := round(-10000000 * d);
      if hTimer<>0 then begin
        IsWaitableTimerEnabled:=true;
        if not SetWaitableTimer(hTimer,t,0,nil,nil,false) then begin // set a timer of 100ns accuracy
          PrintLastError;
          d:=d*0.1;
        end;
        WaitForSingleObjectEx(hTimer,round(d*10000),true);
        IsWaitableTimerEnabled:=false;
      end else begin
        Sleep(round(d*1000)); //  <- filthy hack for old win95
      end;

      dt:=GetTicks;
      if fEnabled and (fInterval>0) then begin
        if Assigned(fOnTimer) then fOnTimer(self);
      end;
    end;

    IsActive:= false;
    suspend;
  end;
end;

function TRealTimer.GetTicks: double;
var n: int64;
begin
  if CpuFreq=0 then result := GetTickCount*0.001
  else begin
    ReaDTimeStampCounter(n);
    result:= n/CpuFreq;
  end;
end;

procedure TRealTimer.SetEnabled(const Value: boolean);
var t: int64;
begin
  if fEnabled = Value then exit;
  fEnabled := Value;
  if fEnabled then begin
    if (fInterval>0) and Suspended and not isActive then Resume;
  end else begin
    if IsWaitableTimerEnabled and not Suspended and IsActive then begin
      t:=-1;
      if hTimer<>0 then if SetWaitableTimer(hTimer,t,0,nil,nil,false) then begin // signal the timer immediately
        while not Suspended and IsActive do begin // let it happen
          Application.ProcessMessages;
          Sleep(10);
        end;
      end;
    end;
  end;
end;

procedure TRealTimer.SetInterval(const Value: double);
var t: int64;
begin
  if fInterval = Value then exit;
  fInterval := Value;
  if not fEnabled then exit;

  if fInterval>0 then begin
    if Suspended and not isActive then Resume;
  end else begin
    if IsWaitableTimerEnabled and not Suspended and IsActive then begin
      t:=-1;
      if hTimer<>0 then SetWaitableTimer(hTimer,t,0,nil,nil,false); // signal the timer immediately
      while not Suspended and IsActive do begin // let it happen
        Sleep(10);
      end;
    end;
  end;
end;
end.
